-- This hash maps WoW chat command tokens to GemSay destination tokens.  Note
-- that CHANNEL is treated differently, because that is used for the
-- rarely-used /c or /csay command (/c 1 message), rarely used because people
-- usually just type "/1 message" instead.

hash_chatToken_to_dest = {
   CHANNEL="CHANNEL",
   EMOTE="e",
   GUILD="g",
   INSTANCE_CHAT="i",
   OFFICER="o",
   PARTY="p",
   RAID="r",
   RAID_WARNING="rw",
   SAY="s",
   YELL="y",
}

-- This is a hash used by GemSay_check_chat_cmd to speed up accesses.

hash_chatDest = {}

-- This is a hash used by GemSay_check_emote to speed up accesses.

hash_emoteToken = {}

function GemSay_initialize()

   --[[ Create /gs and /gsw commands ]]

   SlashCmdList["GS"] = GemSay_gs_command
   SLASH_GS1 = "/gs"
   SlashCmdList["GEMSAY"] = GemSay_gs_command
   SLASH_GEMSAY1 = "/gemsay"
   gemsay_data = {}
   gemsay_sex = {"unknown", "male", "female"}
   gemsay_nom = {"it", "he", "she"}
   gemsay_obj = {"it", "him", "her"}
   gemsay_pos = {"its", "his", "her"}
   gemsay_poss = {"its", "his", "hers"}
   GemSay_quickhelp()
end

function GemSay_error(err)

   --[[ Just put up an error message in a standardized way. ]]

   UIErrorsFrame:AddMessage(err, 0.8, 0.4, 1.0, 1.0, 5)
   PlaySound("igQuestFailed")
end

function GemSay_print(msg)

   --[[ Prints a line to the chat frame. ]]

   DEFAULT_CHAT_FRAME:AddMessage(msg, 0.8, 0.4, 1.0)
end

function GemSay_quickhelp()

   --[[ Prints a summary of commands. ]]

   GemSay_print("GemSay v"..GetAddOnMetadata("GemSay","Version").." by Gemminie of Ner'zhul/Shadowsong <trueflint@gmail.com>")
   GemSay_print("Commands:")
   GemSay_print("/gemsay (/gs) <destination(s)> <message>: Tries to send message to each destination, stopping as soon as it succeeds")
   GemSay_print("/gsr: Open editor window for random strings")
   GemSay_print("/gst: Open editor window for triggered messages")
   GemSay_print("Type '/gs help' for more info.")
end

function GemSay_gs_help()

   --[[ Prints some help. ]]

   GemSay_print("GemSay commands available:")
   GemSay_print("/gemsay (/gs) <dest> <msg>: format <msg> and send to first <dest> possible")
   GemSay_print("<dest> = sequence of these letters, in the order you want to try them:")
   GemSay_print("a = raid warning; d = debug; e = emote; g = guild; i = instance;")
   GemSay_print("o = guild officer chat; p = party; r = raid; s = say; y = yell;")
   GemSay_print("w(player) = whisper player; <number> = numbered chat channel")
   GemSay_print("The /gemsay command (/gs for short) will send the message to the first destination possible.")
   GemSay_print("Certain codes in <msg> will be replaced based on your target or location.")
   GemSay_print("Target: %tt = name; %tr = race; %tc = class; %tl = level; %tf = faction; %ts = sex;")
   GemSay_print("%tn = he/she/it; %to = him/her/it; %tp = his/her/its; %tq = his/hers/its")
   GemSay_print("%TN, %TO, %TP, %TQ, and %TS = capitalized %tn, %to, %tp, %tq, and %ts.")
   GemSay_print("Focus: like target, %ft, %fr, etc.")
   GemSay_print("You: %m = map coords; %w = minimap text; %x = real zone name;")
   GemSay_print("%y = subzone name; %z = zone name; %g<category> = random string")
   GemSay_print("%^tr changes race to ALL CAPS; %_tr changes to all lowercase; works for any code but %g")
   GemSay_print("To open the random string category editor, type /gsr.")
   GemSay_print("To open the message trigger editor, type /gst.")
end

function GemSay_capitalize(str)

   --[[ Just capitalizes the first letter of a string. ]]

   return gsub(str, "^%l", strupper)
end

function GemSay_splitfirst(string)

   --[[ Just split off the first word of the string ]]

   local left, right, first, rest

   if string == "" then
      return "", ""
   end

   left, right = strfind(string, "%S+")	-- first sequence of nonblanks
   if not left then			-- string must be all blanks
      return "", ""
   end
   first = strsub(string, left, right)	-- save the first word
   string = strsub(string, right + 1)	-- now look at the rest of the string
   left, right = strfind(string, "%S+")	-- next sequence of nonblanks
   if not left then			-- must be no more nonblanks
      return first, ""
   end
   rest = strsub(string, left)		-- first nonblank to end of string
   return first, rest
end

function GemSay_matchstring(str, pattern)

   --[[ This little function just calls strfind, and if strfind returns
        a true result (non-nil), returns the string that matched.  If
        there was no match, returns nil. ]]

   local match, len = strfind(str, pattern)

   if match then
      return strsub(str, match, len)
   else
      return nil
   end
end

function GemSay_idiot(msg)

   --[[ Prevent the player from looking like an idiot, even if he/she is one.
   What this means: If there's a "targety" %-code in the string, but
   there's no target, or if there's a "petty" %-code but no pet, or if
   there's a "focusy" %-code but no focus, return true.  If there's
   nothing wrong, return false.  This would be if there's a "targety"
   %-code and there's a target selected, etc., or if there's no "targety"
   %-code. ]]

  local status = false -- initially assume nothing wrong

  if strfind(strlower(msg), "%%[_^]?e[cflnopqrst]") then
    if not UnitExists("pet") then
      GemSay_error("Message refers to pet (%e), but you have no pet; saying nothing")
      status = true
    end
  end
  if strfind(strlower(msg), "%%[_^]?f[cflnopqrst]") then
    if not UnitExists("focus") then
      GemSay_error("Message refers to focus (%f), but you have no focus; saying nothing")
      status = true -- there's a focusy %f code but no focus
    end
  end
  if strfind(strlower(msg), "%%[_^]?t[cflnopqrst]") then
    if not UnitExists("target") then
      GemSay_error("Message refers to target (%t), but you have no target; saying nothing")
      status = true -- there's a targety %t code but no target
    end
  end
  return status
end

function GemSay_random_string(cat)

  --[[ Looks at the number of strings in gemsay_global and gemsay_char in the
       given category, picks a random one, and returns it. ]]

  if cat == "" then return "" end

  local max_global, max_char, random

  -- First, how many strings are there for this category in each table?

  if gemsay_global and gemsay_global[cat] then
    max_global = table.getn(gemsay_global[cat])
  else
    max_global = 0
  end
  if gemsay_char and gemsay_char[cat] then
    max_char = table.getn(gemsay_char[cat])
  else
    max_char = 0
  end

  -- If there aren't any, there aren't any.  Oh well.

  if (max_global + max_char) == 0 then return "" end

  -- If there's at least 1, pick a random one.

  random = math.random(max_global + max_char);

  if random > max_global then
    return gemsay_char[cat][random - max_global];
  else
    return gemsay_global[cat][random];
  end
end

function GemSay_catrecurse(msg, seen_cat)

  --[[ Substitutes a random string from the indicated category for any %g<cat>
       codes, unless they're in seen_cat, in which case they get hidden.
       Guarantees there will be no more %g codes in the result.  Replaces
       %g<cat> until no further %g<cat> codes remain. ]]

  local new_cat = {}
  local hide_cat = {}

  --[[ Get rid of "pathological" cases where we have a %g followed by a
       non-allowed character, or a %g at the end of the string ]]

  msg = gsub(msg, "%%g;?([^_%w])", "%1")
  msg = gsub(msg, "%%g$", "")

  --[[ Jump out if we have no %g codes at all, meaning either we had none to
       begin with, or those last two gsubs got rid of them. ]]

  if not strfind(msg, "%%g") then return msg end

  --[[ If we are here, there must be at least one %g code, and it must be
       followed by an underscore or alphanumeric -- in other words, a category
       name! ]]

  --[[ Inventory of cats!  Make a list of the cats found in this string that
       haven't been seen before, and a list of the ones that have.  This avoids
       changing the string while we're iterating over a gmatch on it. ]]

  for cat in string.gmatch(msg, "%%g([_%w]+)") do

    -- Do not change msg in this loop!

    if seen_cat[cat] then
      hide_cat[cat] = true		-- mark this cat for hiding
    else
      new_cat[cat] = true		-- we haven't seen this cat before
    end
  end

  --[[ The seen-before cats will hide now ]]

  for cat in pairs(hide_cat) do
    msg = gsub(msg, "%%g"..cat..";?", "")
  end

  --[[ Now, for each new cat, replace each occurrence with a random string from
       that cat, after running that string through this function again ]]

  for cat in pairs(new_cat) do
    local new_seen = seen_cat

    new_seen[cat] = true
    msg = gsub(msg, "%%g("..cat..");?",
               function(x)
                 return GemSay_catrecurse(GemSay_random_string(x), new_seen)
               end)
  end

  return msg
end

function GemSay_catsub(msg)

  --[[ Pretty much just serves as an initial seed for GemSay_catrecurse. ]]

  local seen_cat = {}

  -- All we do is seed the recursive function

  return GemSay_catrecurse(msg, seen_cat)
end

function GemSay_targetcode(unit, code, args)

  --[[ Return the meaning of the given % code -- unit is "target", "player",
       "pet", etc.  The "%" and all prefixes are expected to have been removed
       from the code by this time.  Only target-related codes come here. ]]

  local value = ""

  if unit == "***name***" then
    if code == "t" then
      return args[4]
    end
  end

  if code == "c" then
    if UnitIsPlayer(unit) then
      value = UnitClass(unit)
    else
      value = "NPC"
    end
  elseif code == "f" then
    local x

    x, value = UnitFactionGroup(unit)
    if not value then value = "unaffiliated" end
  elseif code == "l" then
    local classification = UnitClassification(unit)

    value = UnitLevel(unit)
    if value == -1 then value = "??" end
    if classification == "worldboss" then
      value = value.."++"
    elseif strsub(classification, -5) == "elite" then
      value = value.."+"
    end
    if strsub(classification, 1, 4) == "rare" then
      value = value.." Rare"
    end
  elseif code == "n" then
    value = gemsay_nom[UnitSex(unit)]
  elseif code == "o" then
    value = gemsay_obj[UnitSex(unit)]
  elseif code == "p" then
    value = gemsay_pos[UnitSex(unit)]
  elseif code == "q" then
    value = gemsay_poss[UnitSex(unit)]
  elseif code == "r" then
    value = UnitRace(unit) or UnitCreatureFamily(unit) or UnitCreatureType(unit) or "unknown"
  elseif code == "s" then
    value = gemsay_sex[UnitSex(unit)]
  elseif code == "t" then
    value = UnitName(unit)
  end
  return value
end

function GemSay_nontargetcode(code, args)

  --[[ Return the meaning of the given % code -- args is the array of event
       arguments, if we were called by an event as opposed to a command or
       macro.  The "%" and all prefixes are expected to have been removed from
       the code by this time.  Only non-target-related codes come here. ]]

  local value = ""

  if code == "m" then
    local x, y = GetPlayerMapPosition("player")

    value = format("(%d, %d)", x*100, y*100)
  elseif code == "w" then
    value = GetMinimapZoneText()
  elseif code == "x" then
    value = GetRealZoneText()
  elseif code == "y" then
    value = GetSubZoneText()
  elseif code == "z" then
    value = GetZoneText()
  elseif strfind(code, "^[1-9]$") then
    local n = tonumber(code)

    if args and args[n] then
      value = args[n]
    end
  end
  return value
end

function GemSay_construct_targetstring(prefixes)

  -- Construct a "targetfocustarget" type string out of a prefix string.  For
  -- example, "tetf" changes to "targetpettargetfocus".  Keep in mind that this
  -- may construct impossibilities.

  local result = ""
  local i

  for i = 1, strlen(prefixes) do
    local char = strsub(prefixes, i, i)

    if char == "e" then result = result .. "pet"
    elseif char == "f" then result = result .. "focus"
    elseif char == "p" then result = result .. "player"
    elseif char == "t" or char == "s" then result = result .. "target" end
  end
  return result
end

function GemSay_getcode(targety, code, args)

  -- If the code is targety (true/false), call GemSay_targetcode.  If not,
  -- _nontargetcode.  In either case, figure out whether we're capitalizing,
  -- all-capsing, lowercasing, or none of the above, and if it's targety,
  -- whether we're referring to "target", "focus", "player", or "pet".  The %
  -- has been removed from code by this time.

  local infocode = strsub(code, -1)
  local prefixes = strlower(strsub(code, 1, -2))
  local upordown = GemSay_matchstring(prefixes, "[_^]")
  local upper = (upordown == "^")
  local lower = (upordown == "_")
  local whichtarget = GemSay_matchstring(prefixes, "[efpst]+")
  local unit = "target"
  local capitalize = false
  local result = ""

  code = strsub(code, -1)	-- Should only get one code letter

  if strfind(infocode, "[A-Z]") then
    capitalize = true
  end

  -- Go through the prefixes and find the ultimate target, if it makes sense.
  -- For example, ttf = target's target's focus.

  -- However, there is one very special unit code.  If the only unit code is
  -- "s", then we use args[4], which is hacked to be the target of the last
  -- spell cast, for situations where you are casting at your focus or by name.
  -- This only works for %st.

  if whichtarget then
    if whichtarget == "s" and args ~= nil then
      unit = "***name***"	-- special value
    else
      unit = GemSay_construct_targetstring(whichtarget)
    end
  end

  code = strlower(code)

  if targety then
    result = GemSay_targetcode(unit, code, args)
  else
    result = GemSay_nontargetcode(code, args)
  end

  if upper then
    return strupper(result)
  elseif lower then
    return strlower(result)
  elseif capitalize then
    return GemSay_capitalize(result)
  else
    return result
  end    
end

function GemSay_nontargetsub(msg, args)

  --[[ Substitutes numerous bits of non-target-related information into the
       given string. ]]

  return gsub(msg, "%%([_^]*[mwxyz1-9])",
              function(x) return GemSay_getcode(false, x, args) end)
end

function GemSay_targetsub(msg, args)

  --[[ Substitutes numerous bits of target-related information into the given
       string. ]]

  return gsub(msg, "%%([_^efpst]*[cflnNoOpPqQrsSt])",
              function(x) return GemSay_getcode(true, x, args) end)
end

function GemSay_idiot_free_expand(msg, args)

  -- Expands all the %g codes out in msg.  If idiocy occurs, returns nil.  The
  -- "args" argument is an array of the event arguments, if this was called
  -- from an event.  If we weren't called by an event, then args will be nil.
  -- If we were called by an event that has no arguments, then args will be an
  -- empty array.

  -- Do random categories -- these strings can contain target-related % codes,
  -- so we need to expand them before checking

  msg = GemSay_catsub(msg)

  -- Prevent potential idiocy

  if (GemSay_idiot(msg)) then
    return nil
  end

  -- Do rest of substitutions

  msg = GemSay_nontargetsub(msg, args)
  msg = GemSay_targetsub(msg, args)
  gemsay_data = {}
  return msg
end

function GemSay_check_chat_cmd(msg)
   -- Check to see whether msg is a chat command.  The way to do this is to see
   -- whether there is a global (these are from FrameXML/GlobalStrings.lua)
   -- named SLASH_xj (where x is one of the keys of hash_chatToken_to_dest and
   -- j is an integer from 1 to ... however many there are) equal to the
   -- proposed slash command.  As we go, we'll build a hash, hash_chatDest,
   -- that makes this shorter, and we can check it before doing the search.

   -- BTW, if this turns out to be a chat command, return the GemSay
   -- destination code for it.  If it isn't, return nil.

   -- First, if there's no slash to start with, forget it.  It isn't a chat
   -- command.

   if not strfind(msg, "^/%w") then return nil end

   -- Now split off the first word.

   local command = strmatch(msg, "^(/%w+)") or nil

   -- Make sure we don't have a situation like msg == "/", "/ ", etc.

   if not command then return nil end

   -- Break off the rest, if there is a rest.

   local remainder = ""
   if command ~= msg then
      remainder = strsub(msg, strlen(command) + 2)
   end

   -- If the command is a number (as in a chat channel), deal with that.

   if strfind(command, "^/%d+") then
      local channel = strmatch(command, "^/(%d+)")
      return channel
   end

   -- Put command in all lowercase for standardization.

   command = strlower(command)

   -- See if we have a hash entry for that command.

   if hash_chatDest[command] then
      return hash_chatDest[command]
   end

   -- If there isn't a hash entry, then we have to look it up, which takes a
   -- lot of trouble, so if we find it, put it in the hash to save trouble
   -- later.

   for token, tokendest in pairs(hash_chatToken_to_dest) do
      local i = 1
      local globalstr = nil
      repeat
	 globalstr = getglobal("SLASH_"..token..i)
	 if globalstr then
	 end
	 if globalstr and strlower(globalstr) == strlower(command) then
	    local dest = tokendest
	    hash_chatDest[command] = dest
	    return dest
	 end
	 i = i + 1
      until not globalstr
   end
   return nil
end

function GemSay_check_emote(msg)
   -- Check and see if msg is an emote.  If it is, return its emote token.  If
   -- it isn't, return nil.  The way to do this is to use the EMOTEi_CMDj
   -- globals (from FrameXML/GlobalStrings.lua), which give the slash commands
   -- for the emotes, and the EMOTEi_TOKEN globals (from
   -- FrameXML/ChatFrame.lua), which have the same index number (the "i" that
   -- I'm using here) and give the tokens.  Like the GemSay_check_chat_cmd function,
   -- we're going to use a hash to speed up repeated accesses.

   -- First of all, if there's no / at the beginning, it can't be an emote.

   if not strfind(msg, "/%w") then return nil end

   -- The first word of the check is all we care about.  If it turns out that
   -- the first word after the / is an emote command, completely ignore the
   -- rest.

   local command = strmatch(msg, "^(/%w+)") or nil

   -- It's not an emote if it's "/" or "/ " or the like.

   if not command then return nil end

   -- Put command in all lowercase for standardization.

   command = strlower(command)

   -- See if command is listed in the hash.

   if hash_emoteToken[command] then
      return hash_emoteToken[command]
   end

   -- Loop through the emotes until we find the right one.

   local globalstr = nil
   local i = 1

   -- The EMOTEi_CMDj globals are defined in GlobalStrings.lua.  Their
   -- values look like "/agree" and the like.  The value of i never exceeds
   -- MAXEMOTEINDEX.  The value of j depends on how many synonyms are
   -- defined for EMOTEi.

   repeat
      globalstr = nil
      local j = 1
      repeat
	 globalstr = getglobal("EMOTE"..i.."_CMD"..j)
	 if globalstr and strlower(globalstr) == command then
	    
	    -- We found it!  Find the token and record it in the hash.
	    
	    local token = getglobal("EMOTE"..i.."_TOKEN")
	    hash_emoteToken[command] = token
	    return token
	 end
	 j = j + 1
      until not globalstr
      i = i + 1
      globalstr = getglobal("EMOTE"..i.."_CMD1")
   until not globalstr
   return nil
end

function GemSay_guild_officer_speak()
   -- Returns true if the character is in a guild and can speak in officer
   -- chat.  Returns false if they can't, or if they're not in a guild.

  if not IsInGuild() then return false end
  local _, _, rankindex = GetGuildInfo("player")
  GuildControlSetRank(rankindex)
  local _, _, _, ospeakflag = GuildControlGetRankFlags()
  return ospeakflag
end

function GemSay_say_one(dest, msg)
   -- Check whether it's possible to send the message to the given
   -- single-character destination code, and if so, send the message and return
   -- true.  Otherwise, send nothing and return false.

   -- First check for standard /-codes for other types of messages and, if
   -- found, change the destination to those types and remove the /-code from
   -- the msg.  Logical recursion is fun.  Obviously this can't work for the
   -- "debug" target since there's no standard Blizzard code for that, and
   -- obviously if you've got a message starting with "/p" and you're not in a
   -- party, nothing's going to show up.  Also, I'm not going to try tackling
   -- "/w" at the moment, as that seems rather tricky.  This is an expansion on
   -- a suggestion from Yltseirp of Cairne.

   local destcheck = GemSay_check_chat_cmd(msg)
   if destcheck then
      dest = destcheck
      msg = strmatch(msg, "^/%w+%s+(.*)$")
      if dest == "CHANNEL" then
	 dest, msg = strmatch(msg, "^(%d+)%s+(.*)$")
      end
   else
      local emotecheck = GemSay_check_emote(msg)
      if emotecheck then

	 -- We have an emote.  If there's anything after the emote in the
	 -- message, issue a false second argument to DoEmote, the effect of
	 -- which is to heed the target if there is one.  If there is no
	 -- target, or if there is nothing after the emote in the message, give
	 -- DoEmote a true second argument, which means to behave as if there
	 -- is no target.

	 local ignoretarget = true
	 if strfind(msg, "^/%w+%s+%w") then
	    ignoretarget = false
	 end
	 DoEmote(emotecheck, ignoretarget)
	 return true
      end
   end

   -- Now, it either wasn't a "/" chat or emote command, or isn't anymore

   -- Now, based on the value of dest, send the message, performing at least
   -- some checks to make sure it's possible.

--   if IsInGroup(LE_PARTY_CATEGORY_HOME) then
--      GemSay_print("IsInGroup(LE_PARTY_CATEGORY_HOME): true")
--   else
--      GemSay_print("IsInGroup(LE_PARTY_CATEGORY_HOME): false")
--   end
--   if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
--      GemSay_print("IsInGroup(LE_PARTY_CATEGORY_INSTANCE): true")
--   else
--      GemSay_print("IsInGroup(LE_PARTY_CATEGORY_INSTANCE): false")
--   end
--   if IsInRaid() then
--      GemSay_print("IsInRaid(): true")
--   else
--      GemSay_print("IsInRaid(): true")
--   end
   if dest == "s" then
      SendChatMessage(msg, "SAY")
      return true
   elseif dest == "e" then
      SendChatMessage(msg, "EMOTE")
      return true
   elseif dest == "y" then
      SendChatMessage(msg, "YELL")
      return true
   elseif dest == "d" then
      GemSay_print(msg)
      return true
   elseif dest == "p" then
      if IsInGroup(LE_PARTY_CATEGORY_HOME) then
	 SendChatMessage(msg, "PARTY")
	 return true
      else
	 return false
      end
   elseif dest == "r" then
      if IsInRaid() then
	 SendChatMessage(msg, "RAID")
	 return true
      else
	 return false
      end
   elseif dest == "a" then
      if IsInRaid() then
	 SendChatMessage(msg, "RAID_WARNING")
	 return true
      else
	 return false
      end
   elseif dest == "g" then
      if IsInGuild() then
	 SendChatMessage(msg, "GUILD")
	 return true
      else
	 return false
      end
   elseif dest == "i" then
      if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
	 SendChatMessage(msg, "INSTANCE_CHAT")
	 return true
      else
	 return false
      end
   elseif dest == "o" then
      if GemSay_guild_officer_speak() then
	 SendChatMessage(msg, "OFFICER")
	 return true
      else
	 return false
      end
   elseif tonumber(dest) then
      if GetChannelName(tonumber(dest)) then
	 SendChatMessage(msg, "CHANNEL", nil, tonumber(dest))
	 return true
      else
	 return false
      end
   end
   return false
end

function GemSay_say_whisper(dest, msg)

   -- See if there's a "w(x)" construction and, if so, send the message to x
   -- and return true.  Otherwise, return false.

   if strlen(dest) >= 4 then -- make sure there's room for at least a "w(x)"
      local pstart, pend, wname = strfind(dest, "w%((%a+)%)")
      if pstart == nil then
	 return false
      else
	 SendChatMessage(msg, "WHISPER", nil, wname)
	 return true
      end
   else -- there was a "w" but no room for "(x)" at the end; we're done
      return false
   end
end

function GemSay_say_stuff(dest, msg, args)

   -- Iterate through the characters of dest, calling GemSay_say_one on each
   -- one and exiting as soon as one succeeds.  Include idiot-prevention
   -- measures.

   -- Expand all % codes, and watch for idiocy

   msg = GemSay_idiot_free_expand(msg, args)

   if not msg then return end

   -- Decide who to say the message to

   local len = strlen(dest)
   local i
   for i = 1, len do
      local char = strsub(dest, i, i)
      if char == "w" then
	 if GemSay_say_whisper(strsub(dest, i, -1), msg) then
	    break
	 end
      end
      if GemSay_say_one(char, msg) then
	 break
      end
   end
end

function GemSay_gs_command(line, args)

  --[[ Break off the destination string from the message and send to
       GemSay_say_stuff.  The "args" array is a list of event arguments, not
       command-line arguments. ]]

  local dest, msg = GemSay_splitfirst(line)

  if dest == "" then
    GemSay_quickhelp()
    return
  end

  if strlower(dest) == "help" then
    GemSay_gs_help()
    return
  end

  if strlower(dest) == "random" then
    GemSay_gsr_command()
    return
  end

  if strlower(dest) == "trigger" then
    GemSay_gst_command()
    return
  end

  if msg == "" then
    GemSay_error("No message given")
    return
  end

  GemSay_say_stuff(dest, msg, args)
end
